/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 2001 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#include <stdio.h>     
#include <stdlib.h>     
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#ifdef WITH_DATABASE

#include "samhain.h"

#include "sh_cat.h"
#include "sh_error.h"
#include "sh_utils.h"

#undef  FIL__
#define FIL__  _("sh_database.c")

typedef struct my_attr_ 
{
  char * attr;
  char * attr_o;
  int    val;
  int    size;
  size_t off;
} my_attr;

typedef struct dbins_ {
  struct dbins_ * next;
  char            host[64];
  char            time[16];
  char            date[16];
  char            msg[1024];
  char            sev[8];
  char            path[1024];
  char            user[9];
  char            group[9];
  char            program[8];
  char            subroutine[16];
  char            status[12];
  char            hash[50];
  char            path_data[1024];
  char            hash_data[50];
  char            key_uid[64];
  char            key_uid_data[64];
  char            key_id[16];
  char            module[8];
  int   	  return_code;
  char            syscall[16];
  char            ip[16];     
  char            tty[16];    
  char            peer[64];   
  char            obj[1024];   
  char            interface[64];   
  char            ltime[64];   
  char            dir[1024];   
  char            linked_path[1024]; 
  int             port;
  char            service[64];   
  char            facility[32];   
  char            priority[32];   
  char            syslog_msg[1024];

  char            mode_old[16];
  char            mode_new[16];
  char            device_old[16];
  char            device_new[16];
  char            owner_old[9];
  char            owner_new[9];
  char            group_old[9];
  char            group_new[9];
  char            ctime_old[20];
  char            ctime_new[20];
  char            atime_old[20];
  char            atime_new[20];
  char            mtime_old[20];
  char            mtime_new[20];
  char            chksum_old[50];
  char            chksum_new[50];
  char            link_old[1024];
  char            link_new[1024];

  long            size_old;
  long            size_new;
  long            hardlinks_old;
  long            hardlinks_new;
  long            inode_old;
  long            inode_new;

} dbins;

static my_attr attr_tab[] = {
  { NULL, N_("sev"),           1,    8, offsetof(struct dbins_, sev) },
  { NULL, N_("tstamp"),        2,   16, offsetof(struct dbins_, time) },
  { NULL, N_("host"),          3,   64, offsetof(struct dbins_, host) },
  { NULL, N_("msg"),           4, 1024, offsetof(struct dbins_, msg) },

  { NULL, N_("path"),          5, 1024, offsetof(struct dbins_, path)  },
  { NULL, N_("user"),          6,    9, offsetof(struct dbins_, user)  },
  { NULL, N_("group"),         7,    9, offsetof(struct dbins_, group)  },
  { NULL, N_("program"),       8,    8, offsetof(struct dbins_, program)  },
  { NULL, N_("subroutine"),    9,   16, offsetof(struct dbins_, subroutine)},
  { NULL, N_("status"),       10,   12, offsetof(struct dbins_, status)  },
  { NULL, N_("hash"),         11,   50, offsetof(struct dbins_, hash)  },
  { NULL, N_("path_data"),    12, 1024, offsetof(struct dbins_, path_data)  },
  { NULL, N_("hash_data"),    13,   50, offsetof(struct dbins_, hash_data)  },
  { NULL, N_("key_uid"),      14,   64, offsetof(struct dbins_, key_uid)  },
  { NULL, N_("key_uid_data"), 15,   64, offsetof(struct dbins_, key_uid_data)},
  { NULL, N_("key_id"),       16,   16, offsetof(struct dbins_, key_id)  },
  { NULL, N_("module"),       17,    8, offsetof(struct dbins_, module)  },
  { NULL, N_("return_code"),  18,    0, offsetof(struct dbins_, return_code)},
  { NULL, N_("syscall"),      19,   16, offsetof(struct dbins_, syscall)  },
  { NULL, N_("ip"),           20,   16, offsetof(struct dbins_, ip)  },
  { NULL, N_("tty"),          21,   16, offsetof(struct dbins_, tty)  },
  { NULL, N_("peer"),         22,   64, offsetof(struct dbins_, peer)  },
  { NULL, N_("obj"),          23, 1024, offsetof(struct dbins_, obj)  },
  { NULL, N_("interface"),    24,   64, offsetof(struct dbins_, interface)  },
  { NULL, N_("time"),         25,   64, offsetof(struct dbins_, ltime)  },
  { NULL, N_("dir"),          26, 1024, offsetof(struct dbins_, dir)  },
  { NULL, N_("linked_path"),  27, 1024, offsetof(struct dbins_, linked_path)},
  { NULL, N_("port"),         28,    0, offsetof(struct dbins_, port)  },
  { NULL, N_("service"),      29,   64, offsetof(struct dbins_, service)  },
  { NULL, N_("facility"),     30,   32, offsetof(struct dbins_, facility)  },
  { NULL, N_("priority"),     31,   32, offsetof(struct dbins_, priority)  },
  { NULL, N_("syslog_msg"),   32, 1024, offsetof(struct dbins_, syslog_msg)  },

  { NULL, N_("mode_old"),     33,   16, offsetof(struct dbins_, mode_old)  },
  { NULL, N_("mode_new"),     34,   16, offsetof(struct dbins_, mode_new)  },
  { NULL, N_("device_old"),   35,   16, offsetof(struct dbins_, device_old)}, 
  { NULL, N_("device_new"),   36,   16, offsetof(struct dbins_, device_new)  },
  { NULL, N_("owner_old"),    37,    9, offsetof(struct dbins_, owner_old)  },
  { NULL, N_("owner_new"),    38,    9, offsetof(struct dbins_, owner_new)  },
  { NULL, N_("group_old"),    39,    9, offsetof(struct dbins_, group_old)  }, 
  { NULL, N_("group_new"),    40,    9, offsetof(struct dbins_, group_new)  }, 
  { NULL, N_("ctime_old"),    41,   20, offsetof(struct dbins_, ctime_old)  }, 
  { NULL, N_("ctime_new"),    42,   20, offsetof(struct dbins_, ctime_new)  }, 
  { NULL, N_("atime_old"),    43,   20, offsetof(struct dbins_, atime_old)  }, 
  { NULL, N_("atime_new"),    44,   20, offsetof(struct dbins_, atime_new)  }, 
  { NULL, N_("mtime_old"),    45,   20, offsetof(struct dbins_, mtime_old)  }, 
  { NULL, N_("mtime_new"),    46,   20, offsetof(struct dbins_, mtime_new)  }, 
  { NULL, N_("chksum_old"),   47,   50, offsetof(struct dbins_, chksum_old)  },
  { NULL, N_("chksum_new"),   48,   50, offsetof(struct dbins_, chksum_new)  },
  { NULL, N_("link_old"),     49, 1024, offsetof(struct dbins_, link_old)  },  
  { NULL, N_("link_new"),     50, 1024, offsetof(struct dbins_, link_new)  },
               
  { NULL, N_("size_old"),     51,    0, offsetof(struct dbins_, size_old)  },
  { NULL, N_("size_new"),     52,    0, offsetof(struct dbins_, size_new)  },
  { NULL, N_("hardlinks_old"),53,    0, offsetof(struct dbins_, hardlinks_old)},
  { NULL, N_("hardlinks_new"),54,    0, offsetof(struct dbins_, hardlinks_new)},
  { NULL, N_("inode_old"),    55,    0, offsetof(struct dbins_, inode_old)  }, 
  { NULL, N_("inode_new"),    56,    0, offsetof(struct dbins_, inode_new)  }, 

  { NULL, NULL,       0, 0, 0 }
};

static char db_name[64]     = "samhain"; 
static char db_table[64]    = "log"; 

static char db_host[64]     = "localhost"; 
static char db_user[64]     = ""; 
static char db_password[64] = "";

static int insert_value (char * ptr, char * str)
{
  if (!ptr || !str)
    return -1;
  if (strlen(str) > 63)
    return -1;
  sl_strlcpy(ptr, str, 64);
  return 0;
}

static void init_db_entry (dbins * ptr)
{
  memset (ptr, '\0', sizeof(dbins));
  ptr->next = NULL;
  return;
}
  

int sh_database_set_database (char * str)
{
  return insert_value (db_name, str);
}
int sh_database_set_table (char * str)
{
  return insert_value (db_table, str);
}
int sh_database_set_host (char * str)
{
  return insert_value (db_host, str);
}
int sh_database_set_user (char * str)
{
  return insert_value (db_user, str);
}
int sh_database_set_password (char * str)
{
  return insert_value (db_password, str);
}

#ifdef WITH_POSTGRES
#include "libpq-fe.h"
int sh_database_query (char  * query, long * id)
{
  char *pghost, *pgport, *pgoptions, *pgtty;
  char* dbName;
  char * p;
  PGconn* conn;
  PGresult* res;

  *id       = 0; /* does PG have auto_increment indexes ? 
		  * How can one access the current index value ?
		  */ 
  pghost    = db_host[0] == '\0'     ? NULL : db_host;
  pgport    = NULL;
  pgoptions = NULL;
  pgtty     = NULL;
  dbName    = db_name[0] == '\0'     ? NULL : db_name;

  conn = PQsetdb(pghost, pgport, pgoptions, pgtty, dbName);

  if (PQstatus(conn) == CONNECTION_BAD) 
    goto err_out;

  res = PGexec(conn, query);
  if (PQresultStatus(res) != PGRES_COMMAND_OK) 
    {
      PQclear(res);
      goto err_out;
    }
  PQclear(res);
  PQfinish(conn);
  SL_RETURN(0, _("sh_database_query"));


 err_out:
  p = PQerrorMessage(conn);
  sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN,
		  p, _("sh_database_query"));
  PQfinish(conn);
  SL_RETURN(status, _("sh_database_query"));
}
#endif 


#ifdef WITH_MYSQL 
#include "mysql.h"
int sh_database_query (char  * query, long * id)
{
  char *     p;
  int        status = 0;
  MYSQL      db_conn;

  SL_ENTER(_("sh_database_query"));

  if (NULL == mysql_init(&db_conn))
    {
	p      = mysql_error (&db_conn);
	status = mysql_errno (&db_conn);
	goto err_out;
    }

  status = 0;
  
  if (NULL == mysql_real_connect(&db_conn, 
				 db_host[0] == '\0'     ? NULL : db_host, 
				 db_user[0] == '\0'     ? NULL : db_user, 
				 db_password[0] == '\0' ? NULL : db_password,
				 db_name[0] == '\0'     ? NULL : db_name, 
				 0, NULL, 0))
    {
      p      = mysql_error (&db_conn);
      status = mysql_errno (&db_conn);
      goto err_out;
    }
  
  if (0 != mysql_query(&db_conn, query))
    {
      status = mysql_errno (&db_conn);
      p      = mysql_error (&db_conn);
      goto err_out;
    }

  p = mysql_info (&db_conn);
  sh_error_handle(SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN,
		  p, _("sh_database_query"));

  *id = (long) mysql_insert_id(&db_conn);

  mysql_close(&db_conn);
  SL_RETURN(0, _("sh_database_query"));

 err_out:
  *id = 0;
  sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_SUBGEN,
		  p, _("sh_database_query"));
  mysql_close(&db_conn);
  SL_RETURN(status, _("sh_database_query"));
}
#endif

char * null_or_val (char * end, char * val, int * size)
{
  int len;

  if (!end  || !val || !size)
    return NULL;

  if (val[0] == '\0')
    {
      sl_strlcat(end, _(",NULL"), *size);
      end   += 5;
      *size -= 5;
    }
  else
    {
      if (*size > 1)
	{
	  *end = ','; ++end; *end = '"'; ++end; (*size) -= 2; *end = '\0';
	}
      len = strlen(val);
      if (*size > (len+1))
	{
	  sl_strlcat(end, val, *size); (*size) -= len;
	  end   += len; *end = '"'; ++end; *end = '\0'; 
	}
    }
  return end;
}

long sh_database_entry (dbins * db_entry, long id)
{
  char   query[1024];
  int    status;
  long   the_id;
  int    size;
  char * end;
  char * p;
  int    i;
  char   num[64];

  SL_ENTER(_("sh_database_entry"));

  if (db_entry->host[0] == '\0')
    {
      if (sh.host.name[0] == '\0')
	strcpy (db_entry->host, _("localhost"));         /* known to fit  */
      else
	sl_strlcpy (db_entry->host, sh.host.name, 64); 
    }

  if (id > 0)
    sprintf(num, "%ld", id);

  sl_snprintf (query, 1023,
	       _("INSERT INTO %s VALUES (NULL,%s,\"%s\",\"%s\",\"%s\",\"%s\""),
	       db_table,
	       id > 0 ? num : _("NULL"),
	       db_entry->host,
	       db_entry->time,
	       db_entry->date,
	       db_entry->sev);

  size =  1023 - strlen(query);
  end  = query + strlen(query);

  i = 3;

  while (attr_tab[i].attr != NULL)
    {
      if (attr_tab[i].size != 0)
	{
	  if (attr_tab[i].val > 40 && attr_tab[i].val < 47)
	    {
	      p = (char *)(db_entry)+attr_tab[i].off;
	      p = strchr(p, 'T');
	      if (p) *p = ' ';
	    }
	  end = null_or_val (end, ((char *)(db_entry)+attr_tab[i].off), &size);
	}
      else if (attr_tab[i].val == 18)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->return_code);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 28)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->port);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 51)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->size_old);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 52)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->size_new);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 53)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->hardlinks_old);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 54)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->hardlinks_new);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 55)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->inode_old);
	  while (*end != '\0') { ++end; --size; }
	}
      else if (attr_tab[i].val == 56)
	{
	  sl_snprintf(end, (size-1), _(",\"%d\""), db_entry->inode_new);
	  while (*end != '\0') { ++end; --size; }
	}
      ++i;
    }

  if (size > 1) { *end = ')'; ++end; *end = '\0'; }

  status = sh_database_query (query, &the_id);

  SL_RETURN(the_id, _("sh_database_entry"));
}

/* this is not a real XML parser, but it copes with the XML format of
 * the log messages provided by sh_error_handle()
 */
char *  sh_database_parse (char * message, dbins * db_entry)
{
  static  int first = S_TRUE;
#ifdef SH_STEALTH
  int     k;
#endif 
  char  * p;
  char  * q;
  dbins * new;
  int     i, j;

  SL_ENTER(_("sh_database_parse"));

  if (!message || *message == '\0')
    SL_RETURN (NULL, _("sh_database_parse"));

#ifdef SH_STEALTH
  if (first == S_TRUE)
    {
      i = 0;
      while (attr_tab[i].attr_o != NULL)
	{
	  j = strlen(attr_tab[i].attr_o);
	  attr_tab[i].attr = sl_malloc (j+1);
	  if (NULL == attr_tab[i].attr)
	    SL_RETURN (NULL, _("sh_database_parse"));
	  for (k = 0; k < j; ++k)
	    attr_tab[i].attr[k] = attr_tab[i].attr_o[k] ^ XOR_CODE;
	  attr_tab[i].attr[j] = '\0';
	  ++i;
	}
      first = S_FALSE;
    }
#else
  if (first == S_TRUE)
    {
      i = 0;
      while (attr_tab[i].attr_o != NULL)
	{
	  attr_tab[i].attr = attr_tab[i].attr_o;
	  ++i;
	}
      first = S_FALSE;
    }
#endif

  p = strchr (message, '<');
  if (!p)
    SL_RETURN (NULL, _("sh_database_parse"));

  while (p && *p != '\0' && *p != '>')
    {
      if (p[0] == '/' && p[1] == '>')
	SL_RETURN (&p[2], _("sh_database_parse"));
      if (p[0] == '/' && p[1] == 'l' && p[2] == 'o' && 
	  p[3] == 'g' && p[4] == '>')
	SL_RETURN (&p[5], _("sh_database_parse"));
      if (p[0] == 'l' && p[1] == 'o' && p[2] == 'g' &&
	  (p[3] == ' ' || p[3] == '>'))
	{
	  p = &p[4];
	  goto parse;
	}
      ++p;
    }
  SL_RETURN(NULL, _("sh_database_parse")); 

 parse:

  while (p && (*p == ' ' || *p == '>'))
    ++p;

  if (!p || *p == '\0')
    SL_RETURN(NULL, _("sh_database_parse"));

  if (p[0] == '<' && p[1] == 'l' &&
      p[2] == 'o' && p[3] == 'g')
    {
      new       = SH_ALLOC(sizeof(dbins));
      init_db_entry(new);
      db_entry->next = new;
      p = sh_database_parse (p, new);
    }

  if (p[0] == '/' && p[1] == '>')
    SL_RETURN (&p[1], _("sh_database_parse"));
  
  if (p[0] == '<' && p[1] == '/' && p[2] == 'l' &&
      p[3] == 'o' && p[4] == 'g' && p[5] == '>')
    SL_RETURN (&p[5], _("sh_database_parse"));


  /* non-whitespace 
   */
  i = 0;
  while (attr_tab[i].attr != NULL)
    {
      j = strlen(attr_tab[i].attr);
      if (0 == strncmp(p, attr_tab[i].attr, j) && 
	  p[j] == '=' && p[j+1] == '"')
	{
	  q = strchr(&p[j+2], '"');
	  if (!q)
	    SL_RETURN(NULL, _("sh_database_parse"));
	  else
	    {
	      *q = '\0';

	      if      (attr_tab[i].val == 1)
		sl_strlcpy(db_entry->sev, &p[j+2], attr_tab[i].size);
	      else if (attr_tab[i].val == 2)
		{
		  sl_strlcpy(db_entry->date, &p[j+2],  11);
		  sl_strlcpy(db_entry->time, &p[j+13],  9);
		}
	      else if (attr_tab[i].val == 3)
		sl_strlcpy(db_entry->host, &p[j+2], attr_tab[i].size);
	      else if (attr_tab[i].val == 4)
		sl_strlcpy(db_entry->msg,  &p[j+2], attr_tab[i].size);
	      else if (attr_tab[i].size != 0)
		{
		  sl_strlcpy( (((char *)(db_entry))+ attr_tab[i].off),
			      &p[j+2], 
			      attr_tab[i].size);
		}
	      else if (attr_tab[i].val == 18)
		{
		  db_entry->return_code = atoi(&p[j+2]);
		}
	      else if (attr_tab[i].val == 28)
		{
		  db_entry->port = atoi(&p[j+2]);
		}

	      *q = '"';
	      p  = q; 
	      ++p;

	      goto parse;
	    }
	}
      ++i;
    }

  /* unknown attribute, skip
   */
  while (p && *p != ' ')
    ++p;

  goto parse;
}

int sh_database_insert (char * message)
{
  dbins * db_entry;
  dbins * prev;
  dbins * curr;
  long    id = 0;

  SL_ENTER(_("sh_database_insert"));

  db_entry        = SH_ALLOC(sizeof(dbins));
  init_db_entry(db_entry);

  /* recursively parse the message into a linked list
   */
  sh_database_parse (message, db_entry);

  /* enter the list into the database
   */
  curr = db_entry;
  if (curr->next)
    {
      prev = curr->next;
      id = sh_database_entry (prev, 0);
      SH_FREE(prev);
    }

  sh_database_entry (curr, id);
  SH_FREE(curr);

  SL_RETURN(0, _("sh_database_insert"));
}

#endif
